'use strict'

var test = require('tap').test
var http = require('http')
var koa = require('koa')
var pinoLogger = require('./')
var split = require('split2')

function setup (t, middleware, cb) {
  var app = koa()
  app.silent = true
  app.use(middleware)
  var server = app.listen(0, '127.0.0.1', function (err) {
    cb(err, server)
  })
  app.use(function * (next) {
    if (this.request.url === '/') {
      this.body = 'hello world'
    }
    yield next
  })
  t.tearDown(function (cb) {
    server.close(cb)
  })

  return app
}

function doGet (server) {
  var address = server.address()
  http.get('http://' + address.address + ':' + address.port)
}

test('default settings', function (t) {
  var dest = split(JSON.parse)
  var logger = pinoLogger(dest)

  setup(t, logger, function (err, server) {
    t.error(err)
    doGet(server)
  })

  dest.on('data', function (line) {
    t.ok(line.req, 'req is defined')
    t.ok(line.res, 'res is defined')
    t.equal(line.msg, 'request completed', 'message is set')
    t.equal(line.req.method, 'GET', 'method is get')
    t.equal(line.res.statusCode, 200, 'statusCode is 200')
    t.end()
  })
})

test('exposes the internal pino', function (t) {
  t.plan(1)

  var dest = split(JSON.parse)
  var logger = pinoLogger(dest)

  dest.on('data', function (line) {
    t.equal(line.msg, 'hello world')
  })

  logger.logger.info('hello world')
})

test('exposes request bound child logger on context, req, res, request, response objects', function (t) {
  var dest = split(JSON.parse)
  var logger = pinoLogger(dest)
  var app = setup(t, logger, function (err, server) {
    t.error(err)
    doGet(server)
  })

  app.use(function * (next) {
    t.equal(this.req.log, this.log)
    t.equal(this.res.log, this.log)
    t.equal(this.request.log, this.log)
    t.equal(this.response.log, this.log)
    this.log.info('test')
    yield next
  })

  dest.once('data', function (line) {
    t.equal(line.msg, 'test', 'msg should be "test"')
    t.ok(line.req, 'should be child logger with req')
    t.end()
  })
})

test('allocate a unique id to every request', function (t) {
  t.plan(5)

  var dest = split(JSON.parse)
  var logger = pinoLogger(dest)
  var lastId = null

  setup(t, logger, function (err, server) {
    t.error(err)
    doGet(server)
    doGet(server)
  })

  dest.on('data', function (line) {
    t.notEqual(line.req.id, lastId)
    lastId = line.req.id
    t.ok(line.req.id, 'req.id is defined')
  })
})

test('supports errors in the response', function (t) {
  var dest = split(JSON.parse)
  var logger = pinoLogger(dest)

  var app = setup(t, logger, function (err, server) {
    t.error(err)
    var address = server.address()
    http.get('http://' + address.address + ':' + address.port + '/error')
  })

  app.use(function * (next) {
    if (this.request.url === '/error') {
      this.body = ''
      this.res.emit('error', Error('boom!'))
    }
    yield next
  })

  dest.on('data', function (line) {
    t.ok(line.req, 'req is defined')
    t.ok(line.res, 'res is defined')
    t.ok(line.err, 'err is defined')
    t.equal(line.msg, 'request errored', 'message is set')
    t.equal(line.req.method, 'GET', 'method is get')
    t.equal(line.res.statusCode, 200, 'statusCode is 200')
    t.end()
  })
})

test('supports errors in the middleware', function (t) {
  var dest = split(JSON.parse)
  var logger = pinoLogger(dest)

  var app = setup(t, logger, function (err, server) {
    t.error(err)
    var address = server.address()
    http.get('http://' + address.address + ':' + address.port + '/error')
  })

  app.use(function * (next) {
    if (this.request.url === '/error') {
      this.body = ''
      throw Error('boom!')
    }
    yield next
  })

  dest.once('data', function (line) {
    // logging the error:
    t.ok(line.req, 'req is defined')
    t.ok(line.res, 'res is defined')
    t.ok(line.err, 'err is defined')
    t.equal(line.err.message, 'boom!', 'err message is boom!')
    t.equal(line.msg, 'request errored', 'message is request errored')
    dest.once('data', function (line) {
      // logging the 500 response:
      t.ok(line.req, 'req is defined')
      t.ok(line.res, 'res is defined')
      t.notOk(line.err, 'err is not defined')
      t.equal(line.msg, 'request completed', 'message is request completed')
      t.equal(line.req.method, 'GET', 'method is get')
      t.equal(line.res.statusCode, 500, 'statusCode is 500')
      t.end()
    })
  })
})

test('does not inhibit downstream error handling', function (t) {
  var dest = split(JSON.parse)
  var logger = pinoLogger(dest)

  var app = setup(t, logger, function (err, server) {
    t.error(err)
    var address = server.address()
    http.get('http://' + address.address + ':' + address.port + '/error')
  })

  app.use(function * (next) {
    try {
      yield next
    } catch (e) {
      t.ok(e)
      t.equal(e.message, 'boom!')
      t.end()
    }
  })

  app.use(function * (next) {
    if (this.request.url === '/error') {
      this.body = ''
      throw Error('boom!')
    }
    yield next
  })
})

test('responseTime', function (t) {
  var dest = split(JSON.parse)
  var logger = pinoLogger(dest)

  var app = setup(t, logger, function (err, server) {
    t.error(err)
    doGet(server)
  })

  function sleep () {
    return new Promise(function (resolve, reject) {
      setTimeout(resolve, 100)
    })
  }

  app.use(function * (next) {
    yield sleep(100)
    yield next
  })

  dest.once('data', function (line) {
    // let's take into account Node v0.10 is less precise
    t.ok(line.responseTime >= 90, 'responseTime is defined and in ms')
    t.end()
  })
})
